// Copyright 2020 Antrea Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package stats

import (
	"context"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	networkingv1 "k8s.io/api/networking/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes/fake"
	featuregatetesting "k8s.io/component-base/featuregate/testing"

	"github.com/vmware-tanzu/antrea/pkg/apis/controlplane"
	secv1alpha1 "github.com/vmware-tanzu/antrea/pkg/apis/security/v1alpha1"
	statsv1alpha1 "github.com/vmware-tanzu/antrea/pkg/apis/stats/v1alpha1"
	fakeversioned "github.com/vmware-tanzu/antrea/pkg/client/clientset/versioned/fake"
	crdinformers "github.com/vmware-tanzu/antrea/pkg/client/informers/externalversions"
	"github.com/vmware-tanzu/antrea/pkg/features"
)

var (
	np1 = &networkingv1.NetworkPolicy{
		ObjectMeta: metav1.ObjectMeta{Namespace: "foo", Name: "bar", UID: "uid1"},
	}
	np2 = &networkingv1.NetworkPolicy{
		ObjectMeta: metav1.ObjectMeta{Namespace: "foo", Name: "baz", UID: "uid2"},
	}
	cnp1 = &secv1alpha1.ClusterNetworkPolicy{
		ObjectMeta: metav1.ObjectMeta{Namespace: "", Name: "bar", UID: "uid3"},
	}
	cnp2 = &secv1alpha1.ClusterNetworkPolicy{
		ObjectMeta: metav1.ObjectMeta{Namespace: "", Name: "baz", UID: "uid4"},
	}
	anp1 = &secv1alpha1.NetworkPolicy{
		ObjectMeta: metav1.ObjectMeta{Namespace: "foo", Name: "bar", UID: "uid5"},
	}
	anp2 = &secv1alpha1.NetworkPolicy{
		ObjectMeta: metav1.ObjectMeta{Namespace: "foo", Name: "baz", UID: "uid6"},
	}
)

func TestAggregatorCollectListGet(t *testing.T) {
	tests := []struct {
		name                                    string
		summaries                               []*controlplane.NodeStatsSummary
		existingNetworkPolicies                 []runtime.Object
		existingAntreaClusterNetworkPolicies    []runtime.Object
		existingAntreaNetworkPolicies           []runtime.Object
		expectedNetworkPolicyStats              []statsv1alpha1.NetworkPolicyStats
		expectedAntreaClusterNetworkPolicyStats []statsv1alpha1.AntreaClusterNetworkPolicyStats
		expectedAntreaNetworkPolicyStats        []statsv1alpha1.AntreaNetworkPolicyStats
	}{
		{
			name: "multiple Nodes, multiple Policies",
			summaries: []*controlplane.NodeStatsSummary{
				{
					ObjectMeta: metav1.ObjectMeta{
						Name: "node-1",
					},
					NetworkPolicies: []controlplane.NetworkPolicyStats{
						{
							NetworkPolicy: controlplane.NetworkPolicyReference{UID: np1.UID},
							TrafficStats: statsv1alpha1.TrafficStats{
								Bytes:    10,
								Packets:  1,
								Sessions: 1,
							},
						},
					},
					AntreaClusterNetworkPolicies: []controlplane.NetworkPolicyStats{
						{
							NetworkPolicy: controlplane.NetworkPolicyReference{UID: cnp1.UID},
							TrafficStats: statsv1alpha1.TrafficStats{
								Bytes:    20,
								Packets:  5,
								Sessions: 2,
							},
						},
					},
					AntreaNetworkPolicies: []controlplane.NetworkPolicyStats{
						{
							NetworkPolicy: controlplane.NetworkPolicyReference{UID: anp1.UID},
							TrafficStats: statsv1alpha1.TrafficStats{
								Bytes:    20,
								Packets:  5,
								Sessions: 2,
							},
						},
					},
				},
				{
					ObjectMeta: metav1.ObjectMeta{
						Name: "node-2",
					},
					NetworkPolicies: []controlplane.NetworkPolicyStats{
						{
							NetworkPolicy: controlplane.NetworkPolicyReference{UID: np1.UID},
							TrafficStats: statsv1alpha1.TrafficStats{
								Bytes:    20,
								Packets:  10,
								Sessions: 3,
							},
						},
					},
					AntreaClusterNetworkPolicies: []controlplane.NetworkPolicyStats{
						{
							NetworkPolicy: controlplane.NetworkPolicyReference{UID: cnp1.UID},
							TrafficStats: statsv1alpha1.TrafficStats{
								Bytes:    20,
								Packets:  8,
								Sessions: 5,
							},
						},
					},
					AntreaNetworkPolicies: []controlplane.NetworkPolicyStats{
						{
							NetworkPolicy: controlplane.NetworkPolicyReference{UID: anp1.UID},
							TrafficStats: statsv1alpha1.TrafficStats{
								Bytes:    100,
								Packets:  10,
								Sessions: 5,
							},
						},
					},
				},
			},
			existingNetworkPolicies:              []runtime.Object{np1, np2},
			existingAntreaClusterNetworkPolicies: []runtime.Object{cnp1, cnp2},
			existingAntreaNetworkPolicies:        []runtime.Object{anp1, anp2},
			expectedNetworkPolicyStats: []statsv1alpha1.NetworkPolicyStats{
				{
					ObjectMeta: metav1.ObjectMeta{
						Namespace: np1.Namespace,
						Name:      np1.Name,
					},
					TrafficStats: statsv1alpha1.TrafficStats{
						Bytes:    30,
						Packets:  11,
						Sessions: 4,
					},
				},
				{
					ObjectMeta: metav1.ObjectMeta{
						Namespace: np2.Namespace,
						Name:      np2.Name,
					},
					TrafficStats: statsv1alpha1.TrafficStats{
						Bytes:    0,
						Packets:  0,
						Sessions: 0,
					},
				},
			},
			expectedAntreaClusterNetworkPolicyStats: []statsv1alpha1.AntreaClusterNetworkPolicyStats{
				{
					ObjectMeta: metav1.ObjectMeta{
						Name: cnp1.Name,
					},
					TrafficStats: statsv1alpha1.TrafficStats{
						Bytes:    40,
						Packets:  13,
						Sessions: 7,
					},
				},
				{
					ObjectMeta: metav1.ObjectMeta{
						Name: cnp2.Name,
					},
					TrafficStats: statsv1alpha1.TrafficStats{
						Bytes:    0,
						Packets:  0,
						Sessions: 0,
					},
				},
			},
			expectedAntreaNetworkPolicyStats: []statsv1alpha1.AntreaNetworkPolicyStats{
				{
					ObjectMeta: metav1.ObjectMeta{
						Namespace: anp1.Namespace,
						Name:      anp1.Name,
					},
					TrafficStats: statsv1alpha1.TrafficStats{
						Bytes:    120,
						Packets:  15,
						Sessions: 7,
					},
				},
				{
					ObjectMeta: metav1.ObjectMeta{
						Namespace: anp2.Namespace,
						Name:      anp2.Name,
					},
					TrafficStats: statsv1alpha1.TrafficStats{
						Bytes:    0,
						Packets:  0,
						Sessions: 0,
					},
				},
			},
		},
		{
			name: "non existing Policies",
			summaries: []*controlplane.NodeStatsSummary{
				{
					ObjectMeta: metav1.ObjectMeta{
						Name: "node-1",
					},
					NetworkPolicies: []controlplane.NetworkPolicyStats{
						{
							NetworkPolicy: controlplane.NetworkPolicyReference{UID: np1.UID},
							TrafficStats: statsv1alpha1.TrafficStats{
								Bytes:    10,
								Packets:  1,
								Sessions: 1,
							},
						},
					},
					AntreaClusterNetworkPolicies: []controlplane.NetworkPolicyStats{
						{
							NetworkPolicy: controlplane.NetworkPolicyReference{UID: cnp1.UID},
							TrafficStats: statsv1alpha1.TrafficStats{
								Bytes:    20,
								Packets:  8,
								Sessions: 5,
							},
						},
					},
					AntreaNetworkPolicies: []controlplane.NetworkPolicyStats{
						{
							NetworkPolicy: controlplane.NetworkPolicyReference{UID: anp1.UID},
							TrafficStats: statsv1alpha1.TrafficStats{
								Bytes:    20,
								Packets:  8,
								Sessions: 5,
							},
						},
					},
				},
			},
			existingNetworkPolicies:              []runtime.Object{np2},
			existingAntreaClusterNetworkPolicies: []runtime.Object{cnp2},
			existingAntreaNetworkPolicies:        []runtime.Object{anp2},
			expectedNetworkPolicyStats: []statsv1alpha1.NetworkPolicyStats{
				{
					ObjectMeta: metav1.ObjectMeta{
						Name:      np2.Name,
						Namespace: np2.Namespace,
					},
					TrafficStats: statsv1alpha1.TrafficStats{
						Bytes:    0,
						Packets:  0,
						Sessions: 0,
					},
				},
			},
			expectedAntreaClusterNetworkPolicyStats: []statsv1alpha1.AntreaClusterNetworkPolicyStats{
				{
					ObjectMeta: metav1.ObjectMeta{
						Name: cnp2.Name,
					},
					TrafficStats: statsv1alpha1.TrafficStats{
						Bytes:    0,
						Packets:  0,
						Sessions: 0,
					},
				},
			},
			expectedAntreaNetworkPolicyStats: []statsv1alpha1.AntreaNetworkPolicyStats{
				{
					ObjectMeta: metav1.ObjectMeta{
						Name:      anp2.Name,
						Namespace: anp2.Namespace,
					},
					TrafficStats: statsv1alpha1.TrafficStats{
						Bytes:    0,
						Packets:  0,
						Sessions: 0,
					},
				},
			},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			defer featuregatetesting.SetFeatureGateDuringTest(t, features.DefaultFeatureGate, features.AntreaPolicy, true)()

			stopCh := make(chan struct{})
			defer close(stopCh)
			client := fake.NewSimpleClientset(tt.existingNetworkPolicies...)
			informerFactory := informers.NewSharedInformerFactory(client, 12*time.Hour)
			crdClient := fakeversioned.NewSimpleClientset(append(tt.existingAntreaClusterNetworkPolicies, tt.existingAntreaNetworkPolicies...)...)
			crdInformerFactory := crdinformers.NewSharedInformerFactory(crdClient, 12*time.Hour)
			a := NewAggregator(informerFactory.Networking().V1().NetworkPolicies(), crdInformerFactory.Security().V1alpha1().ClusterNetworkPolicies(), crdInformerFactory.Security().V1alpha1().NetworkPolicies())
			informerFactory.Start(stopCh)
			crdInformerFactory.Start(stopCh)
			go a.Run(stopCh)

			for _, nodeSummaries := range tt.summaries {
				a.Collect(nodeSummaries)
			}
			// Wait for all summaries being consumed.
			err := wait.PollImmediate(100*time.Millisecond, time.Second, func() (done bool, err error) {
				return len(a.dataCh) == 0, nil
			})
			require.NoError(t, err)
			require.Equal(t, len(tt.expectedNetworkPolicyStats), len(a.ListNetworkPolicyStats("")))
			for _, stats := range tt.expectedNetworkPolicyStats {
				actualStats, exists := a.GetNetworkPolicyStats(stats.Namespace, stats.Name)
				require.True(t, exists)
				require.Equal(t, stats.TrafficStats, actualStats.TrafficStats)
			}
			assert.Equal(t, len(tt.expectedAntreaClusterNetworkPolicyStats), len(a.ListAntreaClusterNetworkPolicyStats()))
			for _, Stats := range tt.expectedAntreaClusterNetworkPolicyStats {
				actualStats, exists := a.GetAntreaClusterNetworkPolicyStats(Stats.Name)
				require.True(t, exists)
				require.Equal(t, Stats.TrafficStats, actualStats.TrafficStats)
			}
			assert.Equal(t, len(tt.expectedAntreaNetworkPolicyStats), len(a.ListAntreaNetworkPolicyStats("")))
			for _, Stats := range tt.expectedAntreaNetworkPolicyStats {
				actualStats, exists := a.GetAntreaNetworkPolicyStats(Stats.Namespace, Stats.Name)
				require.True(t, exists)
				require.Equal(t, Stats.TrafficStats, actualStats.TrafficStats)
			}
		})
	}
}

func TestDeleteNetworkPolicy(t *testing.T) {
	defer featuregatetesting.SetFeatureGateDuringTest(t, features.DefaultFeatureGate, features.AntreaPolicy, true)()

	stopCh := make(chan struct{})
	defer close(stopCh)
	client := fake.NewSimpleClientset(np1)
	informerFactory := informers.NewSharedInformerFactory(client, 12*time.Hour)
	crdClient := fakeversioned.NewSimpleClientset(cnp1, anp1)
	crdInformerFactory := crdinformers.NewSharedInformerFactory(crdClient, 12*time.Hour)
	a := NewAggregator(informerFactory.Networking().V1().NetworkPolicies(), crdInformerFactory.Security().V1alpha1().ClusterNetworkPolicies(), crdInformerFactory.Security().V1alpha1().NetworkPolicies())
	informerFactory.Start(stopCh)
	crdInformerFactory.Start(stopCh)
	go a.Run(stopCh)

	a.Collect(&controlplane.NodeStatsSummary{
		ObjectMeta: metav1.ObjectMeta{
			Name: "node-1",
		},
		NetworkPolicies: []controlplane.NetworkPolicyStats{
			{
				NetworkPolicy: controlplane.NetworkPolicyReference{UID: np1.UID},
				TrafficStats: statsv1alpha1.TrafficStats{
					Bytes:    10,
					Packets:  1,
					Sessions: 1,
				},
			},
		},
		AntreaClusterNetworkPolicies: []controlplane.NetworkPolicyStats{
			{
				NetworkPolicy: controlplane.NetworkPolicyReference{UID: cnp1.UID},
				TrafficStats: statsv1alpha1.TrafficStats{
					Bytes:    30,
					Packets:  3,
					Sessions: 3,
				},
			},
		},
		AntreaNetworkPolicies: []controlplane.NetworkPolicyStats{
			{
				NetworkPolicy: controlplane.NetworkPolicyReference{UID: anp1.UID},
				TrafficStats: statsv1alpha1.TrafficStats{
					Bytes:    30,
					Packets:  3,
					Sessions: 3,
				},
			},
		},
	})
	// Wait for all summaries being consumed.
	err := wait.PollImmediate(100*time.Millisecond, time.Second, func() (done bool, err error) {
		return len(a.dataCh) == 0, nil
	})
	require.NoError(t, err)
	require.Equal(t, 1, len(a.ListNetworkPolicyStats("")))
	require.Equal(t, 1, len(a.ListAntreaClusterNetworkPolicyStats()))
	require.Equal(t, 1, len(a.ListAntreaNetworkPolicyStats("")))

	client.NetworkingV1().NetworkPolicies(np1.Namespace).Delete(context.TODO(), np1.Name, metav1.DeleteOptions{})
	crdClient.SecurityV1alpha1().ClusterNetworkPolicies().Delete(context.TODO(), cnp1.Name, metav1.DeleteOptions{})
	crdClient.SecurityV1alpha1().NetworkPolicies(anp1.Namespace).Delete(context.TODO(), anp1.Name, metav1.DeleteOptions{})
	// Event handlers are asynchronous, it's supposed to finish very soon.
	err = wait.PollImmediate(100*time.Millisecond, time.Second, func() (done bool, err error) {
		return len(a.ListNetworkPolicyStats("")) == 0 && len(a.ListAntreaClusterNetworkPolicyStats()) == 0 && len(a.ListAntreaNetworkPolicyStats("")) == 0, nil
	})
	assert.NoError(t, err)
}
